#!/usr/bin/ruby
#
# DNS Balance --- ưŪʬԤʤ DNS 
#
# By: YOKOTA Hiroshi <yokota@netlab.is.tsukuba.ac.jp>

# $Id: dns_balance.rb,v 1.23 2003/01/31 21:50:34 elca Exp $

# DNS Balance ¸ߤѥ̾
if ENV["ROOT"] == nil
  print("ROOT environment required")
  exit(111)
else
  Prefix = ENV["ROOT"]
end
$LOAD_PATH.unshift(Prefix)

require 'socket'
require 'thread'
require 'getopts'

require 'datatype.rb'
require 'multilog.rb'
require 'log_writer.rb'
require 'util.rb'
require 'as_search.rb'

require 'namespace.rb'
require 'addrdb.rb'

#####################################################################
# 桼㳰
class DnsNotImplementedError < StandardError ; end
class DnsTruncatedError      < StandardError ; end
class DnsNoQueryError        < StandardError ; end
class DnsNoMoreResourceError < StandardError ; end

###################################################################
# ؿ

def usage()
  print "Usage: ruby dns_balance.rb [-lh] [-i ipaddr]\n"
  print "       -i ipaddr  listen IP address (default:0.0.0.0)\n"
  print "       -h         help message\n"
  print "       --as       Enable AS namespace\n"
  exit(111)
end

# DNS ѥåȤƤȼΥפȼΥ饹Ф
def parse_packet(packet)
  (number, flags, num_q, ans_rr, ort_rr, add_rr, str) =  packet.unpack("a2 a2 a2 a2 a2 a2 a*")

  if num_q != "\0\1"
    ML.log("Q must be 1")
    return [nil, nil, nil]
  end

  qlen = str.split("\0")[0].length + 1

  (q, q_type, q_class) = str.unpack("a#{qlen} a2 a2")

  return [q, q_type, q_class]
end

# 饤ȤIPɥ쥹֤
def parse_client_addr(str)
  (family, port, ipaddr, pad) = str.unpack("a2 a2 a4 a*")

  return ipaddr
end

# 饤ȤIPɥ쥹ˤäƤѤ
# ֤̾ƤϤޤʪʤ "default" ˤʤ
def select_namespace(addrstr, name)

  netaddrs = [ipstr_to_str(addrstr)] + ip_masklist(ipstr_to_str(addrstr))

  # custom namespace
  netaddrs.each {
    |i|
    if $namespace_db[i]                  != nil &&
	$addr_db[$namespace_db[i]]       != nil &&
	$addr_db[$namespace_db[i]][name] != nil
      return $namespace_db[i]
    end
  }

  # address number namespace
  netaddrs.each {
    |i|
    if $addr_db[i]        != nil &&
	$addr_db[i][name] != nil
      return i
    end
  }

  # AS namespace
  if $OPT_as &&
      # RFC1918 / ץ饤١ȥɥ쥹Ϥɤ AS ˤ°Ƥʤ
      ipstr_to_str(ipstr_mask(addrstr,  8)) != "10.0.0.0"      &&
      ipstr_to_str(ipstr_mask(addrstr, 12)) != "172.16.0.0"    &&
      ipstr_to_str(ipstr_mask(addrstr, 16)) != "192.168.0.0"   &&
      ipstr_to_str(ipstr_mask(addrstr, 21)) != "204.152.184.0" &&
      ipstr_to_str(addrstr                ) != "127.0.0.1"

    as = as_search(ipstr_to_str(addrstr))
    if as                  != nil &&
	$addr_db[as]       != nil &&
	$addr_db[as][name] != nil
      return as
    end
  end

  return "default"
end

# ŤߤĤѿΤɽ
def make_rand_array(namespace, name)
  rnd_max = 0
  rnd_slesh = []

  $addr_db[namespace][name].each {
    |i|
    rnd_max += (10000 - min(10000, i[1])) # badness κͤ 10000
    rnd_slesh.push(rnd_max)
  }

  return [rnd_max, rnd_slesh]
end

# ŤߤĤ
def select_rand_array(namespace, name, size)
  (rnd_max, rnd_slesh) = make_rand_array(namespace, name)

  if rnd_max == 0  # ƤΥۥȤ Badness  10000 ä
    return []
  end

  arr = []
  (0...size).each {
    |i|
    rnd = rand(rnd_max)
    (0...rnd_slesh.size).each {
      |j|
      if rnd <= rnd_slesh[j]
	arr.push(j)
	break
      end
    }
  }

  return arr
end

# ѥåȤå
def check_packet(q, q_type, q_class)
  # ž̵
  if q_type == DnsType::AXFR
    ML.log("AXFR: " + q.dump + ":" + q_type.dump + ":" + q_class.dump)
    raise DnsNotImplementedError
  end

  # IP(UDP) Τ߼դ
  if !(q_class == DnsClass::INET || q_class == DnsClass::ANY)
    ML.log("noIP: " + q.dump + ":" + q_type.dump + ":" + q_class.dump)
    raise DnsNoQueryError
  end

  # ԲĤʸ
  if (q =~ /[()<>@,;:\\\"\.\[\]]/) != nil
    ML.log("char: " + q.dump + ":" + q_type.dump + ":" + q_class.dump)
    raise DnsNoQueryError
  end

  # A/ANY 쥳ɤΤ߼դ
  if q_type != DnsType::A && q_type != DnsType::ANY
    ML.log("noA: " + q.dump + ":" + q_type.dump + ":" + q_class.dump)
    raise DnsNoQueryError
  end

end

######################################################################
# main

srand()
getopts("lh", "i:0.0.0.0", "d:", "as")

usage() if $OPT_h

ML = MultiLog.new
ML.open

ML.log("dir: " + Prefix)
ML.log("start")


#
# ɥ쥹ǡ١ưŪ
#
Thread.start do
  loop do
    if test(?r, Prefix + "/addr") || test(?r, "./addr")
      begin
	load("addr")

	ML.log("reload")
      rescue NameError,SyntaxError,Exception
	ML.log("reload failed")
      end
    end

    #if test(?r, Prefix + "/addr-once") || test(?r, "./addr-once")
    #  Thread.exit 
    #end

    #p $addr_db
    sleep(5*60) # 5 ʬ˹
  end
end

gs = Socket.open(Socket::AF_INET, Socket::SOCK_DGRAM, 0)
gs.bind([Socket::AF_INET, 53, str_to_ipstr($OPT_i), ""].pack("s n a4 a8"))

#
# ᥤ롼
#
loop do
  (packet, client_addr) = gs.recvfrom(512)
  Thread.start do
    $SAFE = 2
    begin
      client = parse_client_addr(client_addr)
      (q, q_type, q_class) = parse_packet(packet)
      check_packet(q, q_type, q_class) # -> NoQuery, NotImpl

      name      = dnsstr_to_str(q).downcase
      namespace = select_namespace(client, name)

      if $addr_db[namespace] == nil || $addr_db[namespace][name] == nil
	raise DnsNoQueryError
      end

      if $addr_db[namespace][name].size > 1  # -> NameError -> NoQuery
	size = 8
      else
	size = 1
      end
      a_array = select_rand_array(namespace, name, size).uniq

      if a_array.size == 0
	raise DnsNoMoreResourceError
      end

      # 
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x84   # answer & authenticated
      r[3] = 0  # no error

      ans_addrs = []
      a_array.each {
	|i|
	addr = $addr_db[namespace][name][i][0]
	ans_addrs.push(addr.pack("CCCC"))   # 

	# TTL 򡣶ᤤ꤬ʬäƤʤ TTL Ĺ
	if (a_array.size == 1)
	  ttl = "\0\0\x0e\x10" # 1
	else
	  ttl = "\0\0\0\x3c"   # 60
	end

	#  եåȤ 0x000c
	r += "\xc0\x0c" + DnsType::A + DnsClass::INET + ttl + "\0\4" + addr.pack("CCCC")
      }

      # ο򥻥å
      r[6..7] = [a_array.size].pack("n")

      # Ĺ᤮
      if r.length > 512
	raise DnsTruncatedError
      end

      status = "ok"

    rescue DnsNotImplementedError
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x80  # answer
      r[2] &= ~0x04 # not authenticated
      r[3] = 0
      r[3] |= 0x04  # not implemented error
      status = "NotImpl"

    rescue DnsTruncatedError
      # Ĺ᤮Ϻäƥե饰ΩƤ
      r = r[0,512]
      r[2] |= 0x02
      status = "Truncated"

    rescue DnsNoQueryError,DnsNoMoreResourceError,NameError,StandardError
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x84  # answer & authenticated
      r[3] = 0
      r[3] |= 0x03  # name error

      status = "NoQuery"

    rescue
      # ˤʤϤ
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x80  # answer
      r[2] &= ~0x04 # not authenticated
      r[3] = 0
      r[3] |= 0x05  # query refused error
      status = "other"
    end

    #print packet.dump, "\n"
    #print r.dump, "\n"
    #p q

    gs.send(r, 0, client_addr)

    logger(ML, client, status, name, namespace, ans_addrs)

  end
end

# end
